package lcsb.mapviewer.annotation.services;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.apache.commons.collections4.ListUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.aop.framework.Advised;
import org.springframework.aop.support.AopUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.annotation.services.annotators.ElementAnnotator;
import lcsb.mapviewer.annotation.services.annotators.IElementAnnotator;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.user.UserAnnotationSchema;
import lcsb.mapviewer.model.user.UserClassAnnotators;
import lcsb.mapviewer.model.user.annotator.AnnotatorConfigParameter;
import lcsb.mapviewer.model.user.annotator.AnnotatorData;
import lcsb.mapviewer.modelutils.map.ClassTreeNode;
import lcsb.mapviewer.modelutils.map.ElementUtils;
import lcsb.mapviewer.persist.dao.user.UserAnnotationSchemaDao;

/**
 * This class annotates all elements in the model.
 * 
 * @author Piotr Gawron
 * 
 */
@Service
@Transactional
public class ModelAnnotator {

  /**
   * Time required for annotating in
   * {@link #performAnnotations(Model, IProgressUpdater)} method.
   */
  private static final double ANNOTATING_RATIO = 0.8;
  /**
   * Time required for copying annotations from other elements in
   * {@link #performAnnotations(Model, IProgressUpdater)} method.
   */
  private static final double COPYING_RATIO = 0.2;
  /**
   * Default class logger.
   */
  private static Logger logger = LogManager.getLogger();
  /**
   * Connector used for accessing data from miriam registry.
   */
  private MiriamConnector miriamConnector;
  /**
   * List of all avaliable {@link ElementAnnotator} objects.
   */
  private List<IElementAnnotator> availableAnnotators;

  /**
   * List of all avaliable {@link ElementAnnotator} objects.
   */
  private List<IElementAnnotator> defaultAnnotators;

  private UserAnnotationSchemaDao userAnnotationSchemaDao;

  /**
   * Constructor
   */
  @Autowired
  public ModelAnnotator(final MiriamConnector miriamConnector, final List<IElementAnnotator> availableAnnotators,
      final UserAnnotationSchemaDao userAnnotationSchemaDao) {
    this.miriamConnector = miriamConnector;
    this.availableAnnotators = availableAnnotators;
    this.userAnnotationSchemaDao = userAnnotationSchemaDao;
    defaultAnnotators = new ArrayList<>();
    for (final IElementAnnotator annotator : availableAnnotators) {
      if (annotator.isDefault()) {
        defaultAnnotators.add(annotator);
      }
    }
  }

  /**
   * Performs all possible and automatic annotations on the model.
   * 
   * @param model
   *          model to update
   * @param progressUpdater
   *          callback function used for updating progress of the function
   */
  public void performAnnotations(final Model model, final IProgressUpdater progressUpdater) {
    performAnnotations(model, progressUpdater, new UserAnnotationSchema());
  }

  /**
   * Performs all possible and automatic annotations on the model.
   * 
   * @param inputAnnotationSchema
   *          information about {@link ElementAnnotator} objects that should be
   *          used for a given classes
   * @param model
   *          model to update
   * @param progressUpdater
   *          callback function used for updating progress of the function
   */
  public void performAnnotations(final Model model, final IProgressUpdater progressUpdater, final UserAnnotationSchema inputAnnotationSchema) {

    UserAnnotationSchema annotationSchema = inputAnnotationSchema;
    if (annotationSchema != null && annotationSchema.getId() > 0) {
      annotationSchema = userAnnotationSchemaDao.getById(annotationSchema.getId());
    }

    progressUpdater.setProgress(0);
    List<Model> models = new ArrayList<>();
    models.add(model);
    models.addAll(model.getSubmodels());
    final double size = models.size();
    double counter = 0;
    for (final Model m : models) {
      final double ratio = counter / size;

      copyAnnotationFromOtherSpecies(m, new IProgressUpdater() {
        @Override
        public void setProgress(final double progress) {
          progressUpdater.setProgress(ratio * IProgressUpdater.MAX_PROGRESS + (progress * COPYING_RATIO) / size);
        }
      });
      annotateModel(m, new IProgressUpdater() {

        @Override
        public void setProgress(final double progress) {
          progressUpdater.setProgress(ratio * IProgressUpdater.MAX_PROGRESS
              + (COPYING_RATIO * IProgressUpdater.MAX_PROGRESS + progress * ANNOTATING_RATIO) / size);
        }
      }, annotationSchema);
      counter++;
    }
  }

  /**
   * Due to some limitation of CellDesigner, some annotations are missing in xml
   * file. Therefore there are two other places where we look for annotations:
   * <ul>
   * <li>notes,</li>
   * <li>other elements with the same name and type.</li>
   * </ul>
   * <p>
   * This method copies data to elements without annotations from other elements
   * with the same name and type.
   * </p>
   *
   * @param model
   *          model in which element annotations are updated
   * @param progressUpdater
   *          callback function that updates the progress of function execution
   */
  protected void copyAnnotationFromOtherSpecies(final Model model, final IProgressUpdater progressUpdater) {
    double counter = 0;
    double amount = model.getElements().size();
    for (final Species element : model.getSpeciesList()) {
      if (element.getMiriamData().size() == 0) {
        List<Element> speciesList = model.getElementsByName(element.getName());
        for (final Element species2 : speciesList) {
          if (species2.getClass().equals(element.getClass()) && species2.getMiriamData().size() > 0
              && element.getMiriamData().size() == 0) {
            for (final MiriamData md : species2.getMiriamData()) {
              element.addMiriamData(new MiriamData(md));
            }
          }
        }
      }
      counter++;
      progressUpdater.setProgress(counter / amount * IProgressUpdater.MAX_PROGRESS);
    }
  }

  /**
   * Annotates all elements in the model using set of {@link ElementAnnotator}
   * given in the param. If the set is empty then all annotators will be used.
   * 
   * @param annotators
   *          this map contains lists of {@link ElementAnnotator} objects that
   *          should be used for a given classes
   * @param annotatorsParams
   *          this map contains lists of {@link AnnotatorConfigParameter}
   *          objects that should be used for a given {@link ElementAnnotator}
   *          class
   * @param model
   *          model where annotation should be updated
   * @param progressUpdater
   *          callback function used to refresh progress of function execution
   */
  protected void annotateModel(final Model model, final IProgressUpdater progressUpdater, final UserAnnotationSchema annotationSchema) {
    progressUpdater.setProgress(0);
    ElementUtils elementUtils = new ElementUtils();

    double counter = 0;
    double amount = model.getElements().size() + model.getReactions().size();

    // annotate all elements
    for (final BioEntity element : model.getBioEntities()) {
      List<AnnotatorData> list = annotationSchema.getAnnotatorsForClass(element.getClass());
      if (list == null) {
        list = getDefaultAnnotators(element.getClass());
      }
      for (final AnnotatorData annotatorData : list) {
        IElementAnnotator elementAnnotator = null;
        try {
          elementAnnotator = getAnnotator(annotatorData);
          elementAnnotator.annotateElement(element, annotatorData);
        } catch (final AnnotatorException e) {
          logger.warn(elementUtils.getElementTag(element, elementAnnotator) + " annotation problem.", e);
        }
      }
      counter++;
      progressUpdater.setProgress(IProgressUpdater.MAX_PROGRESS * counter / amount);
    }
  }

  private IElementAnnotator getAnnotator(final AnnotatorData annotatorData) throws AnnotatorException {
    for (final IElementAnnotator proxiedElementAnnotator : availableAnnotators) {
      IElementAnnotator elementAnnotator = (IElementAnnotator) unwrapProxy(proxiedElementAnnotator);
      if (annotatorData.getAnnotatorClassName().isAssignableFrom(elementAnnotator.getClass())) {
        return elementAnnotator;
      }
    }
    throw new AnnotatorException("Annotator not supported: " + annotatorData.getAnnotatorClassName());
  }

  /**
   * Returns list of default annotators ({@link ElementAnnotator} with
   * {@link ElementAnnotator#isDefault} flag).
   * 
   * @return list of default annotators
   */
  List<IElementAnnotator> getDefaultAnnotators() {
    return defaultAnnotators;
  }

  /**
   * Returns list of default {@link ElementAnnotator} names that are available
   * for class given in the parameter.
   *
   * @param clazz
   *          class for which list of default annotators will be returned
   * @return annotators names for a given class
   */
  public List<AnnotatorData> getDefaultAnnotators(final Class<?> clazz) {
    List<AnnotatorData> result = new ArrayList<>();
    for (final IElementAnnotator annotator : getAvailableDefaultAnnotators(clazz)) {
      result.add(annotator.createAnnotatorData());
    }
    return result;
  }

  /**
   * This method returns list of improper annotations for the model. Improper
   * annotation is the annotation of the wrong type for given type of element.
   * Types of proper annotations are defined in {@link MiriamType} enum.
   * 
   * @param m
   *          model
   * @param updater
   *          updater call back function that update information about progress
   *          of the function
   * @param validAnnotations
   *          map that contains information which {@link MiriamType miriam
   *          types} are valid for which class
   * @return list of improper annotations
   */
  public Collection<ImproperAnnotations> findImproperAnnotations(final Model m, final IProgressUpdater updater,
      final Map<Class<? extends BioEntity>, Set<MiriamType>> validAnnotations) {
    Map<Class<? extends BioEntity>, Set<MiriamType>> annotations = validAnnotations;
    if (annotations == null) {
      logger.warn(LogMarker.IGNORE, "List of valid annotations is missing. Using default.");
      annotations = getDefaultValidClasses();
    }
    updater.setProgress(0.0);
    List<Model> models = new ArrayList<Model>();
    models.add(m);
    models.addAll(m.getSubmodels());
    List<ImproperAnnotations> result = new ArrayList<>();

    double modelSize = models.size();
    double modelCounter = 0;
    for (final Model model : models) {
      double ratio = modelCounter / modelSize;
      double size = model.getElements().size() + model.getReactions().size();
      double counter = 0;
      for (final Element element : model.getElements()) {
        result.addAll(findImproperAnnotations(element, annotations.get(element.getClass())));
        counter++;
        updater.setProgress(
            ratio * IProgressUpdater.MAX_PROGRESS + (counter / size * IProgressUpdater.MAX_PROGRESS) / modelSize);
      }
      for (final Reaction reaction : model.getReactions()) {
        result.addAll(findImproperAnnotations(reaction, annotations.get(reaction.getClass())));
        counter++;
        updater.setProgress(
            ratio * IProgressUpdater.MAX_PROGRESS + (counter / size * IProgressUpdater.MAX_PROGRESS) / modelSize);
      }
      modelCounter++;
    }
    return result;
  }

  /**
   * This method returns list of improper annotations for the element.
   * 
   * @param element
   *          where we look for improper annotations
   * @param validClasses
   *          list of {@link MiriamType miriam types} that are valid for a given
   *          {@link BioEntity}
   * @return list of improper annotations
   */
  protected List<ImproperAnnotations> findImproperAnnotations(final BioEntity element, final Collection<MiriamType> validClasses) {
    List<ImproperAnnotations> result = new ArrayList<>();
    for (final MiriamData md : element.getMiriamData()) {
      boolean valid = false;
      if (validClasses.contains(md.getDataType())) {
        valid = true;
      }
      if (valid && miriamConnector.getUrlString(md) == null) {
        valid = false;
      }
      if (!valid) {
        result.add(new ImproperAnnotations(element, md));
      }
    }
    return result;
  }

  /**
   * Returns list of all elements that miss annotation.
   * 
   * @param m
   *          model where the search is performed
   * @param requestedAnnotations
   *          map that contains information which {@link MiriamType miriam
   *          types} are obligatory for which class
   * 
   * @return list of all elements that miss annotation
   */
  public Collection<ProblematicAnnotation> findMissingAnnotations(final Model m,
      final Map<Class<? extends BioEntity>, Set<MiriamType>> requestedAnnotations) {
    Map<Class<? extends BioEntity>, Set<MiriamType>> annotations = requestedAnnotations;
    if (annotations == null) {
      logger.warn(LogMarker.IGNORE, "List of requested annotations is missing. Using default.");
      annotations = getDefaultRequiredClasses();
    }
    List<ProblematicAnnotation> result = new ArrayList<>();

    List<Model> models = new ArrayList<>();
    models.add(m);
    models.addAll(m.getSubmodels());
    for (final Model model : models) {
      for (final Element alias : model.getElements()) {
        result.addAll(findMissing(alias, annotations.get(alias.getClass())));
      }
      for (final Reaction reaction : model.getReactions()) {
        result.addAll(findMissing(reaction, annotations.get(reaction.getClass())));
      }
    }
    return result;
  }

  /**
   * Checks if {@link BioEntity} is properly annotated and if not return info
   * about missing annotation.
   * 
   * @param element
   *          object to be checked
   * @param requestedAnnotations
   *          at list one of this annotations should appear in the element, if
   *          the list is null then check is skipped, if the list is empty
   *          (size()==0) then method checks if element contains at least one
   *          miriam
   * @return list of containing improper annotations about element
   */
  private List<ProblematicAnnotation> findMissing(final BioEntity element, final Collection<MiriamType> requestedAnnotations) {
    List<ProblematicAnnotation> result = new ArrayList<>();
    // if there are no requested annotations then don't check
    if (requestedAnnotations == null) {
      return result;
      // if there are no requested annotations, but the list was mentioned then
      // we check if at least one annotation exists
    } else if (requestedAnnotations.size() == 0) {
      if (element.getMiriamData().size() == 0) {
        result.add(new MissingAnnotation(element));
      }
      return result;
    }
    // check if at least one annotation match requested list
    for (final MiriamData md : element.getMiriamData()) {
      for (final MiriamType mt : requestedAnnotations) {
        if (mt.equals(md.getDataType())) {
          return result;
        }
      }
    }
    result.add(new MissingRequiredAnnotations(element, requestedAnnotations));
    return result;
  }

  /**
   * 
   * @return {@link #availableAnnotators}
   */
  public List<IElementAnnotator> getAvailableAnnotators() {
    return availableAnnotators;
  }

  /**
   * Returns list of available annotators for a given class type.
   * 
   * @param clazz
   *          class type
   * @return list of available annotators for a given class type
   */
  public List<IElementAnnotator> getAvailableAnnotators(final Class<?> clazz) {
    List<IElementAnnotator> result = new ArrayList<>();
    for (final IElementAnnotator annotator : availableAnnotators) {
      if (annotator.isAnnotatable(clazz)) {
        result.add(annotator);
      }
    }
    return result;
  }

  /**
   * Returns list of default {@link ElementAnnotator annotators} that are
   * available for class given in the parameter.
   * 
   * @param clazz
   *          class for which list of default annotators will be returned
   * @return annotators for a given class
   */
  List<IElementAnnotator> getAvailableDefaultAnnotators(final Class<?> clazz) {
    List<IElementAnnotator> result = new ArrayList<>();
    for (final IElementAnnotator annotator : availableAnnotators) {
      if (annotator.isAnnotatable(clazz) && annotator.isDefault()) {
        result.add(annotator);
      }
    }
    return result;
  }

  /**
   * Converts list of strings into list of {@link ElementAnnotator}. Strings
   * must be valid {@link ElementAnnotator} common names.
   * 
   * @param list
   *          list of {@link ElementAnnotator#getCommonName()}.
   * @return list of {@link ElementAnnotator annotators}
   */
  public List<IElementAnnotator> getAnnotatorsFromCommonNames(final List<AnnotatorData> list) {
    List<IElementAnnotator> result = new ArrayList<>();
    for (final AnnotatorData annotatorData : list) {
      boolean added = false;
      for (final IElementAnnotator annotator : availableAnnotators) {
        if (annotatorData.getAnnotatorClassName().isAssignableFrom(unwrapProxy(annotator).getClass())) {
          added = true;
          result.add(annotator);
        }
      }
      if (!added) {
        throw new InvalidArgumentException("Unknown annotator name: " + annotatorData.getAnnotatorClassName());
      }
    }
    return result;
  }

  /**
   * Returns map with informations about default valid {@link MiriamType miriam
   * types } for {@link BioEntity} class type.
   * 
   * @return map with informations about valid {@link MiriamType miriam types }
   *         for {@link BioEntity} class type
   */
  @SuppressWarnings("unchecked")
  public Map<Class<? extends BioEntity>, Set<MiriamType>> getDefaultValidClasses() {
    Map<Class<? extends BioEntity>, Set<MiriamType>> result = new HashMap<Class<? extends BioEntity>, Set<MiriamType>>();
    ElementUtils eu = new ElementUtils();
    ClassTreeNode tree = eu.getAnnotatedElementClassTree();

    Queue<ClassTreeNode> nodes = new LinkedList<ClassTreeNode>();
    nodes.add(tree);
    while (!nodes.isEmpty()) {
      ClassTreeNode node = nodes.poll();
      Set<MiriamType> set = new HashSet<MiriamType>();
      Class<? extends BioEntity> clazz = (Class<? extends BioEntity>) node.getClazz();
      for (final MiriamType mt : MiriamType.values()) {
        for (final Class<?> clazz2 : mt.getValidClass()) {
          if (clazz2.isAssignableFrom(clazz)) {
            set.add(mt);
          }
        }
      }
      result.put(clazz, set);
      for (final ClassTreeNode child : node.getChildren()) {
        nodes.add(child);
      }
    }
    return result;
  }

  /**
   * Returns map with informations about default required {@link MiriamType
   * miriam types } for {@link BioEntity} class type.
   * 
   * @return map with informations about required {@link MiriamType miriam types
   *         } for {@link BioEntity} class type
   */
  @SuppressWarnings("unchecked")
  public Map<Class<? extends BioEntity>, Set<MiriamType>> getDefaultRequiredClasses() {
    Map<Class<? extends BioEntity>, Set<MiriamType>> result = new HashMap<>();
    ElementUtils eu = new ElementUtils();
    ClassTreeNode tree = eu.getAnnotatedElementClassTree();

    Queue<ClassTreeNode> nodes = new LinkedList<>();
    nodes.add(tree);
    while (!nodes.isEmpty()) {
      ClassTreeNode node = nodes.poll();
      Set<MiriamType> set = null;
      Class<? extends BioEntity> clazz = (Class<? extends BioEntity>) node.getClazz();
      if ((Boolean) (node.getData())) {
        set = new HashSet<>();
        for (final MiriamType mt : MiriamType.values()) {
          for (final Class<?> clazz2 : mt.getRequiredClass()) {
            if (clazz2.isAssignableFrom(clazz)) {
              set.add(mt);
            }
          }
        }
      }
      result.put(clazz, set);
      for (final ClassTreeNode child : node.getChildren()) {
        nodes.add(child);
      }
    }
    return result;
  }

  public UserAnnotationSchema createDefaultAnnotatorSchema() {
    UserAnnotationSchema result = new UserAnnotationSchema();
    ElementUtils elementUtils = new ElementUtils();

    for (final Class<?> clazz : ListUtils.union(elementUtils.getAvailableElementSubclasses(),
        elementUtils.getAvailableReactionSubclasses())) {
      UserClassAnnotators uca = new UserClassAnnotators(clazz);
      uca.setAnnotators(getDefaultAnnotators(clazz));
      result.addClassAnnotator(uca);
    }
    return result;
  }

  public static final Object unwrapProxy(final Object bean) {

    /*
     * If the given object is a proxy, set the return value as the object being
     * proxied, otherwise return the given object.
     */
    if (AopUtils.isAopProxy(bean) && bean instanceof Advised) {

      Advised advised = (Advised) bean;

      try {
        return advised.getTargetSource().getTarget();
      } catch (Exception e) {
        logger.error("Problem with unwrapping spring proxy", e);
        return bean;
      }
    } else {
      return bean;
    }
  }

}
